﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;

namespace com.tomfosdick.NotThreadSafe
{
    class Program
    {
        /// <summary>
        /// We'll be tring to add 1 to this in 100 different threads
        /// </summary>
        static int accumulator = 0;

        /// <summary>
        /// used for random wait periods in the threads
        /// </summary>
        static Random rand = new Random();

        /// <summary>
        /// an array of 100 threads
        /// </summary>
        static Thread[] threads = new Thread[100];

        /// <summary>
        /// Main method
        /// </summary>
        /// <param name="args"></param>
        static void Main(string[] args)
        {
            for (int i = 0; i < 100; i++)
            {
                //create thread with DoAddition as the work it will do
                threads[i] = new Thread(new ParameterizedThreadStart(DoAddition));
                threads[i].IsBackground = true;
                Console.WriteLine(String.Format("Starting Thread {0}", i));
                threads[i].Start();
            }

            //use a simple polling method to see when they're all finished
            //we'll demonstrate a better method later. 
            bool waiting = true;
            do
            {
                Thread.Sleep(100);
                for (int i = 0; i < 100; i++)
                {
                    if (false == threads[i].IsAlive)
                    {
                        waiting = false;
                        break;
                    }
                }
            } while(waiting);
            Console.WriteLine(String.Format("If this were thread safe, this would be 100: [{0}]", accumulator));
            Console.ReadLine();
        }


        /// <summary>
        /// This is the thread function. There will be (more or less) 100
        /// of these running at the same time.
        /// The basic idea is that it simulates some really complex or time consuming operation in
        /// which the value of accumulator is read, worked on and then written.
        /// </summary>
        /// <param name="obj">not used</param>
        public static void DoAddition(object obj)
        {
            int tempVal;
            Thread.Sleep(rand.Next(10));
            
            /*
             * All you have to do to make this thread safe is uncomment this lock statement
             * Try it!
             */
            //lock (threads)
            {
                tempVal = accumulator;
                
                //wastes some time simulating a complex calculation  
                for (double f = double.MaxValue; f > double.Epsilon; f *= 0.5f) ; 

                accumulator = tempVal + 1;
            }
        }

    }
}
